"""Module for smart_open"""
# Standard
from gzip import GzipFile
from pathlib import Path
import warnings
import shutil
# Installed
import boto3
from cloudpathlib import S3Path, AnyPath


def is_s3(path: str or Path or S3Path):
    """Determine if a string points to an s3 location or not.

    Parameters
    ----------
    path : str or Path or S3Path
        Path to determine if it is and s3 location or not.

    Returns
    -------
    : bool
    """

    if isinstance(path, str):
        return path.startswith('s3://')
    if isinstance(path, Path):
        if str(path).startswith('s3://'):
            warnings.warn("Path object appears to contain an S3 path. "
                          "You should use S3Path to refer to S3 object urls.")
        return False
    if isinstance(path, S3Path):
        return True
    raise ValueError(f"Unrecognized path type for {path} ({type(path)})")


def is_gzip(path: str or Path or S3Path):
    """Determine if a string points to an gzip file.

    Parameters
    ----------
    path : str or Path or S3Path
        Path to check.

    Returns
    -------
    : bool
    """
    if isinstance(path, str):
        return path.endswith('.gz')
    return path.name.endswith('.gz')


def smart_open(path: str or Path or S3Path, mode: str = 'rb', enable_gzip: bool = True):
    """
    Open function that can handle local files or files in an S3 bucket. It also
    correctly handles gzip files determined by a `*.gz` extension.

    Parameters
    ----------
    path : str or Path or S3Path
        Path to the file to be opened. Files residing in an s3 bucket must begin
        with "s3://".
    mode: str, Optional
        Optional string specifying the mode in which the file is opened. Defaults
        to 'rb'.
    enable_gzip : bool, Optional
        Flag to specify that `*.gz` files should be opened as a `GzipFile` object.
        Setting this to False is useful when creating the md5sum of a `*.gz` file.
        Defaults to True.

    Returns
    -------
    : filelike object
    """
    def _gzip_wrapper(fileobj):
        """Wrapper around a filelike object that unzips it
        (if it is enabled and if the file object was opened in binary mode).

        Parameters
        ----------
        fileobj : filelike object
            The original (possibly zipped) object

        Returns
        -------
        : GzipFile or filelike object
        """
        if is_gzip(path) and enable_gzip:
            if 'b' not in mode:
                raise IOError(f'Gzip files must be opened in binary (b) mode. Got {mode}.')
            return GzipFile(filename=path, fileobj=fileobj)
        return fileobj

    if isinstance(path, (Path, S3Path)):
        return _gzip_wrapper(path.open(mode=mode))

    # AnyPath is polymorphic to Path and S3Path. Disable false pylint error
    return _gzip_wrapper(AnyPath(path).open(mode=mode))  # pylint: disable=E1101


def smart_copy_file(source_path: str or Path or S3Path, dest_path: str or Path or S3Path):
    """
        Copy function that can handle local files or files in an S3 bucket.
        Returns the path to the newly created file.

        Parameters
        ----------
        source_path : str or Path or S3Path
            Path to the source file to be copied. Files residing in an s3 bucket must begin
            with "s3://".
        dest_path: str or Path or S3Path
            Path to the Destination file to be copied to. Files residing in an s3 bucket
            must begin with "s3://".

        Returns
        -------
        : Path or S3Path
            The path to the newly created file
        """
    if not is_s3(source_path) and not is_s3(dest_path):
        # This is a local copy and uses shutil copy
        local_source_path = Path(source_path)
        local_dest_path = Path(dest_path)

        # Warning if no suffix is used in destination.
        if len(local_dest_path.suffix) == 0:
            warnings.warn(f'You have copied to a location without a file extension.'
                          f'Source location: {local_source_path} to destination:'
                          f'{local_dest_path}.')

        # Returns a PosixPath of the newly created file
        return shutil.copy(source_path, dest_path)

    # Check if either source or destination is remote and allocate remote resources
    s3 = boto3.resource("s3")
    client = boto3.client("s3")

    if is_s3(dest_path) and not is_s3(source_path):
        # This is a local to remote copy and uses S3 upload
        s3_dest_path = S3Path(dest_path)
        local_source_path = Path(source_path)

        # Warning if no suffix is used.
        if len(s3_dest_path.suffix) == 0:
            warnings.warn(f'You have copied a file to S3 without a file extension.'
                          f'Source location: {local_source_path} to S3 location:'
                          f'{s3_dest_path}.')

        # Has no return, but will raise exceptions on problems
        s3.Bucket(s3_dest_path.bucket).upload_file(str(local_source_path), s3_dest_path.key)
        return s3_dest_path

    if is_s3(source_path) and not is_s3(dest_path):
        # This is a remote to local copy and uses S3 download
        s3_source_path = S3Path(source_path)
        local_dest_path = Path(dest_path)

        # Ensure a full destination path including file name is used
        if local_dest_path.is_dir():
            local_dest_path = local_dest_path / s3_source_path.name
            warnings.warn(f'A directory was given as the destination for the smart file '
                          f'copy. This was modified to include a name as follows.'
                          f'Copy from {s3_source_path} to {local_dest_path}.')

        # Warning if no suffix is used.
        if len(local_dest_path.suffix) == 0:
            warnings.warn(f'You have copied a file without a file extension.'
                          f'Source: {s3_source_path} to destination:'
                          f'{local_dest_path}.')

        # Has no return, but will raise exceptions on problems
        s3.Bucket(s3_source_path.bucket).download_file(s3_source_path.key, str(local_dest_path))
        return Path(local_dest_path)

    # This is a remote to remote copy and uses S3 copy
    s3_source_path = S3Path(source_path)
    s3_dest_path = S3Path(dest_path)

    copy_source = {
        'Bucket': s3_source_path.bucket,
        'Key': s3_source_path.key
    }

    # Warning if no suffix is used.
    if len(s3_dest_path.suffix) == 0:
        warnings.warn(f'You have copied a file to S3 without a file extension.'
                      f'Source location: {s3_source_path} to S3 location:'
                      f'{s3_dest_path}.')

    # Has no return, but will raise exceptions
    client.copy(copy_source, s3_dest_path.bucket, s3_dest_path.key)
    return s3_dest_path



"""Module for anything related to package versioning"""
# Standard
from importlib import metadata


def version():
    """Get package version from metadata"""
    return metadata.version('libera_utils')



"""Quality flag definitions"""

from enum import Flag, EnumMeta, Enum
from operator import or_ as _or_
from functools import reduce

#pylint: disable-all
class FrozenFlagMeta(EnumMeta):
    """
    Metaclass that freezes an enum entirely, preventing values from being updated, added, or deleted.
    """
    def __new__(mcs, name, bases, classdict):
        classdict['__frozenenummeta_creating_class__'] = True
        flag = super().__new__(mcs, name, bases, classdict)
        del flag.__frozenenummeta_creating_class__
        return flag

    def __call__(cls, value, names=None, *, module=None, **kwargs):
        if names is None:  # simple value lookup
            return cls.__new__(cls, value)
        enum = Enum._create_(value, names, module=module, **kwargs)
        enum.__class__ = type(cls)
        return enum

    def __setattr__(cls, name, value):
        members = cls.__dict__.get('_member_map_', {})
        if hasattr(cls, '__frozenenummeta_creating_class__') or name in members:
            return super().__setattr__(name, value)
        if hasattr(cls, name):
            msg = "{!r} object attribute {!r} is read-only"
        else:
            msg = "{!r} object has no attribute {!r}"
        raise AttributeError(msg.format(cls.__name__, name))

    def __delattr__(cls, name):
        members = cls.__dict__.get('_member_map_', {})
        if hasattr(cls, '__frozenenummeta_creating_class__') or name in members:
            return super().__delattr__(name)
        if hasattr(cls, name):
            msg = "{!r} object attribute {!r} is read-only"
        else:
            msg = "{!r} object has no attribute {!r}"
        raise AttributeError(msg.format(cls.__name__, name))


class QualityFlag(Flag):
    """
    Subclass of Flag that add a method for decomposing a flag into its individual components
    and a property to return a list of all messages associated with a quality flag
    """
    # Overriding enum module implementation to incorporate patched _decompose method below
    # TODO: Remove this if we upgrade to python 3.9.0a2 or newer
    @classmethod
    def _create_pseudo_member_(cls, value):
        """
        Create a composite member iff value contains only members.
        """
        pseudo_member = cls._value2member_map_.get(value, None)
        if pseudo_member is None:
            # verify all bits are accounted for
            _, extra_flags = _patched_enum_decompose(cls, value)
            if extra_flags:
                raise ValueError("%r is not a valid %s" % (value, cls.__name__))
            # construct a singleton enum pseudo-member
            pseudo_member = object.__new__(cls)
            pseudo_member._name_ = None
            pseudo_member._value_ = value
            # use setdefault in case another thread already created a composite
            # with this value
            pseudo_member = cls._value2member_map_.setdefault(value, pseudo_member)
        return pseudo_member

    # Overriding enum module implementation to incorporate patched _decompose method below
    # TODO: Remove this if we upgrade to python 3.9.0a2 or newer
    def __repr__(self):
        cls = self.__class__
        if self._name_ is not None:
            return '<%s.%s: %r>' % (cls.__name__, self._name_, self._value_)
        members, uncovered = _patched_enum_decompose(cls, self._value_)
        return '<%s.%s: %r>' % (
                cls.__name__,
                '|'.join([str(m._name_ or m._value_) for m in members]),
                self._value_,
                )

    # Overriding enum module implementation to incorporate patched _decompose method below
    # TODO: Remove this if we upgrade to python 3.9.0a2 or newer
    def __str__(self):
        cls = self.__class__
        if self._name_ is not None:
            return '%s.%s' % (cls.__name__, self._name_)
        members, uncovered = _patched_enum_decompose(cls, self._value_)
        if len(members) == 1 and members[0]._name_ is None:
            return '%s.%r' % (cls.__name__, members[0]._value_)
        else:
            return '%s.%s' % (
                    cls.__name__,
                    '|'.join([str(m._name_ or m._value_) for m in members]),
                    )

    # Overriding enum module implementation to incorporate patched _decompose method below
    # TODO: Remove this if we upgrade to python 3.9.0a2 or newer
    def __invert__(self):
        members, uncovered = _patched_enum_decompose(self.__class__, self._value_)
        inverted = self.__class__(0)
        for m in self.__class__:
            if m not in members and not (m._value_ & self._value_):
                inverted = inverted | m
        return self.__class__(inverted)

    def decompose(self):
        """
        Return the set of all set flags that form a subset of the queried flag value. Note that this is not the
        minimum set of quality flags but rather a full set of all flags such that when they are ORed together, they
        produce self.value

        :return: members, not_covered
            `members` is a list of flag values that are subsets of `value`
            `not_covered` is zero if the OR of members recreates `value`. Non-zero otherwise if bits are set in `value`
                that do not exist as named values in cls.
        """
        value = self.value
        not_covered = value
        flags_to_check = [
            (m, v)
            for v, m in list(self.__class__._value2member_map_.items())
            if m.name is not None
        ]
        members = []
        for member, member_value in flags_to_check:
            if member_value and member_value & value == member_value:
                members.append(member)
                not_covered &= ~member_value
        if not members and value in self.__class__._value2member_map_:
            members.append(self.__class__._value2member_map_[value])
        members.sort(key=lambda m: m._value_, reverse=True)
        return members, not_covered

    @property
    def summary(self):
        members, not_covered = self.decompose()

        if not_covered:
            raise ValueError(f"{self.__name__} has value {self.value} but that value cannot be created by elements "
                             f"of {self.__class__}. This should never happen unless a quality flag was declared "
                             f"without using the FrozenFlagMeta metaclass.")

        try:
            return int(self.value), [m.value.message for m in members]
        except Exception as err:
            raise AttributeError(
                "Tried to summarize a quality flag but its values don't appear to have messages.") from err


class FlagBit(int):
    """Subclass of int to capture both an integer value and an accompanying message"""
    def __new__(cls, *args, message=None, **kwargs):
        obj = super().__new__(cls, *args, **kwargs)
        obj.message = message
        return obj

    def __str__(self):
        return f"{super().__str__()}: {self.message}"


# This is a patched version of enum._decompose method that was fixed in Python 3.9.0a2
# See: https://bugs.python.org/issue38045
# TODO: Remove this if we upgrade to python 3.9.0a2 or newer
def _patched_enum_decompose(flag, value):
    """Extract all members from the value."""
    # _decompose is only called if the value is not named
    not_covered = value
    negative = value < 0
    members = []
    for member in flag:
        member_value = member.value
        if member_value and member_value & value == member_value:
            members.append(member)
            not_covered &= ~member_value
    if not negative:
        tmp = not_covered
        while tmp:
            high_bit = tmp.bit_length() - 1
            flag_value = 2 ** high_bit
            if flag_value in flag._value2member_map_:
                members.append(flag._value2member_map_[flag_value])
                not_covered &= ~flag_value
            tmp &= ~flag_value
    if not members and value in flag._value2member_map_:
        members.append(flag._value2member_map_[value])
    members.sort(key=lambda m: m._value_, reverse=True)
    if len(members) > 1 and members[0].value == value:
        # we have the breakdown, don't need the value member itself
        members.pop(0)
    return members, not_covered


def with_all_none(f):
    """Add NONE and ALL psuedo-members to f"""
    f._member_map_['NONE'] = f(FlagBit(0, message="No flags set."))
    f._member_map_['ALL'] = f(reduce(_or_, f))
    return f



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['libera_utils', 'libera_utils.db', 'libera_utils.io']

package_data = \
{'': ['*'],
 'libera_utils': ['data/config.json',
                  'data/config.json',
                  'data/jpss1_geolocation_xtce_v1.xml',
                  'data/jpss1_geolocation_xtce_v1.xml',
                  'data/spice/*']}

install_requires = \
['SQLAlchemy>=1.4,<2.0',
 'astropy>=5.1,<6.0',
 'cloudpathlib[s3]>=0,<1',
 'h5netcdf>=1.1.0,<2.0.0',
 'h5py>=3.3,<4.0',
 'lasp-packets>=2.0,<3.0',
 'numpy>=1.21,<2.0',
 'psycopg2>=2.9,<3.0',
 'requests>=2.26,<3.0',
 'setuptools>=65',
 'spiceypy>=5.0,<6.0',
 'watchtower>=3.0,<4.0',
 'xarray>=2023.1.0,<2024.0.0']

entry_points = \
{'console_scripts': ['libera-utils = libera_utils.cli:main']}

setup_kwargs = {
    'name': 'libera-utils',
    'version': '2.0.1',
    'description': 'Utility package for Libera Science Data Processing and the Libera Science Data Center. Copyright 2022 University of Colorado.',
    'long_description': 'None',
    'author': 'Stephane Beland',
    'author_email': 'stephane.beland@lasp.colorado.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Module for manifest file handling"""
# Standard
from datetime import datetime
import json
import logging
from pathlib import Path
from hashlib import md5
# Installed
from cloudpathlib import S3Path, AnyPath
# Local
from libera_utils.io.smart_open import smart_open
from libera_utils.io.filenaming import ManifestFilename, ManifestType

logger = logging.getLogger(__name__)


class ManifestError(Exception):
    """Generic exception related to manifest file handling"""
    pass


class Manifest:
    """Object representation of a JSON manifest file"""

    __manifest_elements = (
        "manifest_type",
        "files",
        "configuration"
    )

    def __init__(self, manifest_type: ManifestType,
                 files: list, configuration: dict, filename: str = None):
        self.manifest_type = manifest_type
        self.files = files
        self.configuration = configuration
        self.filename = filename

    def _generate_filename(self):
        """Generate a valid manifest filename"""
        mfn = ManifestFilename.from_filename_parts(
            manifest_type=self.manifest_type,
            created_time=datetime.utcnow()
        )
        return mfn

    def to_json_dict(self):
        """Create a dict representation suitable for writing out.

        Returns
        -------
        : dict
        """
        valid_json = json.dumps({
            'manifest_type': self.manifest_type.value,
            'files': self.files,
            'configuration': self.configuration
        })
        return json.loads(valid_json)

    @classmethod
    def from_file(cls, filepath: str or Path or S3Path):
        """Read a manifest file and return a Manifest object (factory method).

        Parameters
        ----------
        filepath : str or Path or S3Path
            Location of manifest file to read.

        Returns
        -------
        : Manifest
        """
        with smart_open(filepath) as manifest_file:
            contents = json.loads(manifest_file.read())
        for element in cls.__manifest_elements:
            if element not in contents:
                raise ManifestError(f"{filepath} is not a valid manifest file. Missing required element {element}.")
        return cls(ManifestType(contents['manifest_type'].upper()),
                   contents['files'],
                   contents['configuration'],
                   filename=filepath)

    def write(self, outpath: str or Path or S3Path, filename: str = None):
        """Write a manifest file from a Manifest object (self).

        Parameters
        ----------
        outpath : str or Path or S3Path
            Directory path to write to (directory being used loosely to refer also to an S3 bucket path).
        filename : str, Optional
            Optional filename. If not provided, the method uses the objects internal filename attribute. If that is
            not set, then a filename is automatically generated.

        Returns
        -------
        : Path or S3Path
        """
        if filename is None:
            if self.filename is None:
                filename = str(self._generate_filename())
            else:
                filename = self.filename

        filepath = AnyPath(outpath) / filename
        with smart_open(filepath, 'x') as manifest_file:
            json.dump(self.to_json_dict(), manifest_file)
        return filepath

    def validate_checksums(self):
        """Validate checksums of listed files"""
        failed_filenames = []
        for record in self.files:
            checksum_expected = record['checksum']
            filename = record['filename']
            # Validate checksums
            with smart_open(filename, 'rb') as fh:
                checksum_calculated = md5(fh.read(), usedforsecurity=False).hexdigest()
                if checksum_expected != checksum_calculated:
                    logger.error(f"Checksum validation for {filename} failed. "
                                 f"Expected {checksum_expected} but got {checksum_calculated}.")
                    failed_filenames.append(str(filename))
        if failed_filenames:
            raise ValueError(f"Files failed checksum validation: {', '.join(failed_filenames)}")

    def add_file_to_manifest(self, file_path):
        """Add a file to the manifest from filename
          Parameters
        ----------
        file_path : str or Path or S3Path
            Directory path to the file to add to the manifest.

        Returns
        -------
        None
        """
        with smart_open(file_path) as fh:
            checksum_calculated = md5(fh.read(), usedforsecurity=False).hexdigest()
        file_structure = {"filename": str(file_path),
                          "checksum": str(checksum_calculated)}
        self.files.append(file_structure)

    def add_desired_time_range(self, start_datetime: datetime, end_datetime: datetime):
        """Add a file to the manifest from filename
          Parameters
        ----------
        start_datetime : datetime
            The desired start time for the range of data in this manifest

        end_datetime : datetime
            The desired end time for the range of data in this manifest

        Returns
        -------
        None
        """
        self.configuration["start_time"] = start_datetime.strftime('%Y-%m-%d:%H:%M:%S')
        self.configuration["end_time"] = end_datetime.strftime('%Y-%m-%d:%H:%M:%S')
        


"""Database module"""
# Standard
from contextlib import contextmanager
import logging
import os
# Installed
from sqlalchemy import create_engine, MetaData
from sqlalchemy.orm import sessionmaker
# Local
from libera_utils.config import config

logger = logging.getLogger(__name__)

Session = sessionmaker(expire_on_commit=False)


class DatabaseException(Exception):
    """Generic database related error"""
    pass


class _DatabaseManager:
    """
    A caching class used to manage database connections.

    This class should never be instantiated directly. Instead, users should use the convenience method
    libera_utils.db.database.getdb, which is an alias for the _DatabaseManager.get factory method.
    """
    _db_manager_cache = {}

    def __init__(self, dbhost: str, dbuser: str, dbpass: str, dbname: str):
        """_DatabaseManager constructor

        Parameters
        ----------
        dbhost : str
            Database host
        dbuser : str
            Database user
        dbpass : str
            Database password for user login
        dbname : str
            Database name
        """
        self.pid = os.getpid()  # Store the PID of the process in which this object was created

        self.database = dbname or config.get('LIBERA_DB_NAME')
        if not self.database:
            raise DatabaseException("Missing database name.")

        self.user = dbuser or config.get('LIBERA_DB_USER')
        if not self.user:
            raise DatabaseException("Missing database user.")

        if dbhost:  # If we were directly passed a truthy value
            self.host = dbhost
        else:  # No truthy value, then get the environment variable (or json default)
            self.host = config.get('LIBERA_DB_HOST')

        if not self.host:  # If we still don't have a host value, set localhost as a best guess
            self.host = 'localhost'

        self.password = dbpass

        self.engine = create_engine(self.url)
        logger.info(f"Initialized {self}")

    def __str__(self):
        return f"_DatabaseManager(user={self.user}, host={self.host}, db={self.database})"

    def __bool__(self):
        return bool(self.engine)

    def __hash__(self):
        # Used for retrieving cached manager objects instead of creating new ones
        return hash((self.pid, self.url))

    @property
    def url(self):
        """JDBC connection string"""
        return self._format_url(self.database, self.user, self.host, self.password)

    @classmethod
    def get(cls, dbname: str = None, dbuser: str = None, dbhost: str = None, dbpass: str = ""):
        """Cache-enabled factory method.

        Retrieve an existing DB manager from the cache if one already exists in the same PID and configuration
        (identified by hash). If no identical manager exists in this process already, create a new one and return it.
        This makes _DatabaseManager safe for use with either forked or spawned processes because we are never
        copying database engines across process boundaries.

        If no parameters are passed, the _DatabaseManager object sources configuration from the environment.

        Parameters
        ----------
        dbname : str, Optional
            Database name. If not provided, LIBERA_DB_NAME environment variable is used.
        dbuser : str, Optional
            Database user. If not provided, LIBERA_DB_USER environment variable is used.
        dbhost : str, Optional
            Database host. If not provided, 'localhost' is used.
        dbpass : str, Optional
            Database password for user. If not provided, PGPASSWORD environment variable or ~/.pgpass file is used.

        Returns
        -------
        : _DatabaseManager
        """
        # Create a new candidate _DatabaseManager to compare hash-wise to the existing cache
        new_db_manager = cls(dbhost, dbuser, dbpass, dbname)
        try:
            return cls._db_manager_cache[hash(new_db_manager)]
        except KeyError:
            cls._db_manager_cache[hash(new_db_manager)] = new_db_manager
            return new_db_manager

    @staticmethod
    def _format_url(database: str, user: str, host: str, password: str = "",  # nosec B107 # CWE-259
                    port: int = 5432, dialect: str = "postgresql"):
        """
        Returns a database connection url given database parameters

        Parameters
        ----------
        database : str
            Name of database to connect to
        user : str
            DB username
        host : str
            Name of host machine
        password : str, Optional
            Password. Passing an empty string results in searching the environment for PGPASSWORD or the .pgpass file.
        port: int, Optional
            Port number. Default is 5432
        dialect : str, Optional
            SQL dialect. Default is Postgres

        Returns
        -------
        : str
            JDBC connection string
        """
        return f"{dialect}://{user}:{password}@{host}:{port}/{database}"

    @contextmanager
    def session(self):
        """Provide a transactional scope around a series of operations."""
        Session.configure(bind=self.engine)

        session = Session()
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    def truncate_product_tables(self):
        """
        Truncates all tables in sdp schema except for flyway_schema_history
        :return:
        """
        if self.host not in ('localhost', 'local-db'):
            raise ValueError(f"Refusing to truncate all tables for database on host {self.host}. "
                             "We only permit this operation for local dev databases on host "
                             "'local-db' or 'localhost'.")
        meta = MetaData(schema='sdp')
        meta.reflect(bind=self.engine)
        for table in reversed(meta.sorted_tables):
            if table.name not in ('flyway_schema_history', ):
                self.engine.execute(table.delete())
